@@ -1,5 +1,5 @@
 /* grecs - Gray's Extensible Configuration System
-   Copyright (C) 2007-2016 Sergey Poznyakoff
+   Copyright (C) 2007-2021 Sergey Poznyakoff
 
    Grecs is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
@@ -133,6 +133,8 @@ main(int argc, char **argv)
 			grecs_parser_options |= GRECS_OPTION_QUOTED_STRING_CONCAT;
 		else if (strcmp(arg, "-stradj") == 0)
 			grecs_parser_options |= GRECS_OPTION_ADJUST_STRING_LOCATIONS;
+		else if (strncmp(arg, "-preproc=", 9) == 0)
+			grecs_preprocessor = arg + 9;
 		else if (arg[0] == '-')
 			usage(progname, stderr, 1);
 		else {
